#include <os.h>
#include <cmath>
#include <cstdint>

#include "gl.h"
#include "fastmath.h"

#define WIDTH 320
#define HEIGHT 240
#define RGB(r,g,b) (((r&0b11111) << 11) | ((g&0b111111) << 5) | (b&0b11111))

#define M(m, y, x) (m.data[x*4+y])
#define P(m, y, x) (m->data[x*4+y])

#define likely(x) __builtin_expect((x),1)
#define unlikely(x) __builtin_expect((x),0)

MATRIX *transformation;
static COLOR color;
static uint16_t *screen;
static Fix<GLFix::precision, uint32_t> *z_buffer;
static TEXTURE *texture;

void *operator new(size_t t)
{
    return malloc(t);
}

void *operator new[](size_t t)
{
    return malloc(t);
}

void init_gl()
{
    init_fastmath();
    transformation = new MATRIX[MATRIX_STACK_SIZE];
    z_buffer = new Fix<GLFix::precision, uint32_t>[WIDTH*HEIGHT];
    glClear(GL_DEPTH_BUFFER_BIT);
    glLoadIdentity();
    color = RGB(0, 0, 0);
}

void uninit_gl()
{
    uninit_fastmath();
    delete[] transformation;
    delete[] z_buffer;
}

static inline void swap(int *x, int *y)
{
	int t = *y;
	*y = *x;
	*x = t;
}

static void mult_mat_mat(MATRIX *mat1, MATRIX *mat2)
{
    GLFix a00 = P(mat1, 0, 0), a01 = P(mat1, 1, 0), a02 = P(mat1, 2, 0), a03 = P(mat1, 3, 0);
    GLFix a10 = P(mat1, 0, 1), a11 = P(mat1, 1, 1), a12 = P(mat1, 2, 1), a13 = P(mat1, 3, 1);
    GLFix a20 = P(mat1, 0, 2), a21 = P(mat1, 1, 2), a22 = P(mat1, 2, 2), a23 = P(mat1, 3, 2);
    GLFix a30 = P(mat1, 0, 3), a31 = P(mat1, 1, 3), a32 = P(mat1, 2, 3), a33 = P(mat1, 3, 3);

    GLFix b00 = P(mat2, 0, 0), b01 = P(mat2, 1, 0), b02 = P(mat2, 2, 0), b03 = P(mat2, 3, 0);
    GLFix b10 = P(mat2, 0, 1), b11 = P(mat2, 1, 1), b12 = P(mat2, 2, 1), b13 = P(mat2, 3, 1);
    GLFix b20 = P(mat2, 0, 2), b21 = P(mat2, 1, 2), b22 = P(mat2, 2, 2), b23 = P(mat2, 3, 2);
    GLFix b30 = P(mat2, 0, 3), b31 = P(mat2, 1, 3), b32 = P(mat2, 2, 3), b33 = P(mat2, 3, 3);

    P(mat1, 0, 0) = b00*a00 + b01*a10 + b02*a20 + b03*a30;
    P(mat1, 1, 0) = b00*a01 + b01*a11 + b02*a21 + b03*a31;
    P(mat1, 2, 0) = b00*a02 + b01*a12 + b02*a22 + b03*a32;
    P(mat1, 3, 0) = b00*a03 + b01*a13 + b02*a23 + b03*a33;
    P(mat1, 0, 1) = b10*a00 + b11*a10 + b12*a20 + b13*a30;
    P(mat1, 1, 1) = b10*a01 + b11*a11 + b12*a21 + b13*a31;
    P(mat1, 2, 1) = b10*a02 + b11*a12 + b12*a22 + b13*a32;
    P(mat1, 3, 1) = b10*a03 + b11*a13 + b12*a23 + b13*a33;
    P(mat1, 0, 2) = b20*a00 + b21*a10 + b22*a20 + b23*a30;
    P(mat1, 1, 2) = b20*a01 + b21*a11 + b22*a21 + b23*a31;
    P(mat1, 2, 2) = b20*a02 + b21*a12 + b22*a22 + b23*a32;
    P(mat1, 3, 2) = b20*a03 + b21*a13 + b22*a23 + b23*a33;
    P(mat1, 0, 3) = b30*a00 + b31*a10 + b32*a20 + b33*a30;
    P(mat1, 1, 3) = b30*a01 + b31*a11 + b32*a21 + b33*a31;
    P(mat1, 2, 3) = b30*a02 + b31*a12 + b32*a22 + b33*a32;
    P(mat1, 3, 3) = b30*a03 + b31*a13 + b32*a23 + b33*a33;
}

static inline void mult_mat_vect_res(MATRIX *mat1, VERTEX *vect, VERTEX *res)
{
    GLFix x = vect->x, y = vect->y, z = vect->z, w = 1;

    res->x = P(mat1, 0, 0)*x + P(mat1, 0, 1)*y + P(mat1, 0, 2)*z + P(mat1, 0, 3)*w;
    res->y = P(mat1, 1, 0)*x + P(mat1, 1, 1)*y + P(mat1, 1, 2)*z + P(mat1, 1, 3)*w;
    res->z = P(mat1, 2, 0)*x + P(mat1, 2, 1)*y + P(mat1, 2, 2)*z + P(mat1, 2, 3)*w;
         /*w = P(mat1, 3, 0)*x + P(mat1, 3, 1)*y + P(mat1, 3, 2)*z + P(mat1, 3, 3)*w;

    res->x /= w;
    res->y /= w;
    res->z /= w;*/

    res->x -= WIDTH/2;
    res->y -= HEIGHT/2;

    //Prevent div through 0 or a small value (takes too long)
    if(res->z < GLFix(1))
        res->z = 1;

    GLFix div = GLFix(NEAR_PLANE)/res->z;
    res->x *= div;
    res->y *= div;
    res->z -= NEAR_PLANE; //void pixel() clips at 0

    res->x += WIDTH/2;
    res->y += HEIGHT/2;

    //Screen has y=0 at the top, OpenGL at the bottom
    res->y = GLFix(HEIGHT-1) - res->y;

    //TODO: Move this somewhere else
    res->u = vect->u * texture->width;
    res->v = vect->v * texture->height;
}

void set_screen(uint16_t* screenBuf)
{
	screen = screenBuf;
}

void rotate_x(GLFix a)
{
    MATRIX rot;

    GLFix sina = fast_sin(a), cosa = fast_cos(a);

    M(rot, 0, 0) = 1;
    M(rot, 1, 1) = cosa;
    M(rot, 2, 1) = -sina;
    M(rot, 1, 2) = sina;
    M(rot, 2, 2) = cosa;
    M(rot, 3, 3) = 1;

    mult_mat_mat(transformation, &rot);
}

void rotate_y(GLFix a)
{
    MATRIX rot;

    GLFix sina = fast_sin(a), cosa = fast_cos(a);
	
    M(rot, 0, 0) = cosa;
    M(rot, 2, 0) = sina;
    M(rot, 1, 1) = 1;
    M(rot, 0, 2) = -sina;
    M(rot, 2, 2) = cosa;
    M(rot, 3, 3) = 1;

    mult_mat_mat(transformation, &rot);
}

void rotate_z(GLFix a)
{
    MATRIX rot;

    GLFix sina = fast_sin(a), cosa = fast_cos(a);
	
    M(rot, 0, 0) = cosa;
    M(rot, 1, 0) = -sina;
    M(rot, 0, 1) = sina;
    M(rot, 1, 1) = cosa;
    M(rot, 2, 2) = 1;
    M(rot, 3, 3) = 1;

    mult_mat_mat(transformation, &rot);
}

void glLoadIdentity()
{
    *transformation = {}; // Copy empty matrix into transformation
    P(transformation, 0, 0) = P(transformation, 1, 1) = P(transformation, 2, 2) = P(transformation, 3, 3) = 1;
}

void glTranslatef(GLFix x, GLFix y, GLFix z)
{
    MATRIX trans;

    M(trans, 0, 0) = M(trans, 1, 1) = M(trans, 2, 2) = M(trans, 3, 3) = 1;
    M(trans, 0, 3) = x;
    M(trans, 1, 3) = y;
    M(trans, 2, 3) = z;

    mult_mat_mat(transformation, &trans);
}

void glScale3f(GLFix x, GLFix y, GLFix z)
{
    MATRIX scale;

    M(scale, 0, 0) = x;
    M(scale, 1, 1) = y;
    M(scale, 2, 2) = z;
    M(scale, 3, 3) = 1;

    mult_mat_mat(transformation, &scale);
}

inline void pixel(int x, int y, GLFix z, COLOR c)
{
    if(x < 0 || y < 0 || x >= WIDTH || y >= HEIGHT)
        return;

    //Seems to make no difference, gcc uses two shifts
    //asm __volatile__("mla %[x], %[y], %[pitch], %[x]" : [x] "+r" (x): [y] "r" (y), [pitch] "r" (WIDTH) : "r4");
    x += y*WIDTH;

    if(z.value < 0 || z_buffer[x].value <= static_cast<uint32_t>(z.value))
        return;

    z_buffer[x].value = z.value;

    screen[x] = c;
}

inline void swap(VERTEX **v1, VERTEX **v2)
{
    VERTEX *temp = *v2;
    *v2 = *v1;
    *v1 = temp;
}

void draw_line_3d(VERTEX *v1, VERTEX *v2, COLOR c)
{
    //TODO: Vertical lines
    if(v2->x < v1->x)
        swap(&v1, &v2);

    GLFix dy = (v2->y - v1->y) / (v2->x - v1->x);
    GLFix dz = (v2->z - v1->z) / (v2->x - v1->x);

    VERTEX current(*v1);

    for(; current.x <= v2->x; current.x += GLFix(1))
    {
        pixel(current.x.round(), current.y.round(), current.z, c);

        current.y += dy;
        current.z += dz;
    }
}

/* Basic scanline algorithm, no perspective correct texture mapping.
   In fact, this function is horribly buggy. */
static void draw_triangle(VERTEX *low, VERTEX *middle, VERTEX *high, COLOR c)
{
    #ifdef TEXTURE_SUPPORT
        (void) c; //c unused
    #endif

    //I don't know why I have to do this..
    high->x = high->x.wholes();
    high->y = high->y.wholes();

    middle->x = middle->x.wholes();
    middle->y = middle->y.wholes();

    low->x = low->x.wholes();
    low->y = low->y.wholes();

    //We don't want div through 0
    if((low->y == middle->y && high->y == middle->y) || (low->x == middle->x && high->x == middle->x))
        return;

    if(middle->y > high->y)
        swap(&middle, &high);

    if(low->y > middle->y)
        swap(&low, &middle);

    if(middle->y > high->y)
        swap(&middle, &high);

    high->y += 1;

    GLFix height_upper = high->y - middle->y;
    GLFix dx_upper = (high->x - middle->x) / height_upper;
    GLFix dz_upper = (high->z - middle->z) / height_upper;

    GLFix height_lower = (middle->y - low->y) + GLFix(1);
    GLFix dx_lower = (middle->x - low->x) / height_lower;
    GLFix dz_lower = (middle->z - low->z) / height_lower;

    GLFix height_far = high->y - low->y;
    GLFix dx_far = (high->x - low->x) / height_far;
    GLFix dz_far = (high->z - low->z) / height_far;

#ifdef TEXTURE_SUPPORT
    GLFix du_upper = (high->u - middle->u) / height_upper;
    GLFix dv_upper = (high->v - middle->v) / height_upper;

    GLFix du_lower = (middle->u - low->u) / height_lower;
    GLFix dv_lower = (middle->v - low->v) / height_lower;

    GLFix du_far = (high->u - low->u) / height_far;
    GLFix dv_far = (high->v - low->v) / height_far;

    GLFix ustart = low->u, uend = low->u + du_lower;
    GLFix vstart = low->v, vend = low->v + dv_lower;
#endif

    GLFix y = low->y, xstart = low->x, zstart = low->z, xend = low->x + dx_lower, zend = low->z + dz_lower;

    for(; y < middle->y && y <= GLFix(HEIGHT); y += GLFix(1))
    {
        GLFix dz = (zend - zstart) / (xend - xstart);

#ifdef TEXTURE_SUPPORT
        GLFix du = (uend - ustart) / (xend - xstart);
        GLFix dv = (vend - vstart) / (xend - xstart);
#endif

        if(xstart <= xend)
        {
            GLFix z = zstart;
#ifdef TEXTURE_SUPPORT
            GLFix u = ustart;
            GLFix v = vstart;
#endif
            for(GLFix x = xstart; x <= xend; x += 1)
            {
                z += dz;

#ifdef TEXTURE_SUPPORT
                u += du;
                v += dv;
                if(u >= texture->width || u < GLFix(0))
                    u = 0;
                if(v >= texture->height || v < GLFix(0))
                    v = 0;

                pixel(x, y, z, texture->bitmap[u.to_int() + v.to_int()*texture->width]);
#else
                pixel(x, y, z, c);
#endif
            }
        }
        else
        {
            GLFix z = zend;
#ifdef TEXTURE_SUPPORT
            GLFix u = uend;
            GLFix v = vend;
#endif
            for(GLFix x = xend; x <= xstart; x += 1)
            {
                z += dz;

#ifdef TEXTURE_SUPPORT
                u += du;
                v += dv;
                if(u >= texture->width || u < GLFix(0))
                    u = 0;
                if(v >= texture->height || v < GLFix(0))
                    v = 0;

                pixel(x, y, z, texture->bitmap[u.to_int() + v.to_int()*texture->width]);
#else
                pixel(x, y, z, c);
#endif
            }
        }

        xstart += dx_far;
        zstart += dz_far;
        xend += dx_lower;
        zend += dz_lower;

#ifdef TEXTURE_SUPPORT
        ustart += du_far;
        vstart += dv_far;
        uend += du_lower;
        vend += dv_lower;
#endif
    }

    for(; y < high->y && y <= GLFix(HEIGHT); y += GLFix(1))
    {
        GLFix dz = (zend - zstart) / (xend - xstart);
#ifdef TEXTURE_SUPPORT
        GLFix du = (uend - ustart) / (xend - xstart);
        GLFix dv = (vend - vstart) / (xend - xstart);
#endif

        if(xstart <= xend)
        {
            GLFix z = zstart;
#ifdef TEXTURE_SUPPORT
            GLFix u = ustart;
            GLFix v = vstart;
#endif
            for(GLFix x = xstart; x <= xend; x += 1)
            {
                z += dz;

#ifdef TEXTURE_SUPPORT
                u += du;
                v += dv;
                if(u >= texture->width || u < GLFix(0))
                    u = 0;
                if(v >= texture->height || v < GLFix(0))
                    v = 0;

                pixel(x, y, z, texture->bitmap[u.to_int() + v.to_int()*texture->width]);
#else
                pixel(x, y, z, c);
#endif
            }
        }
        else
        {
            GLFix z = zend;
#ifdef TEXTURE_SUPPORT
            GLFix u = uend;
            GLFix v = vend;
#endif
            for(GLFix x = xend; x <= xstart; x += 1)
            {
                z += dz;

#ifdef TEXTURE_SUPPORT
                u += du;
                v += dv;
                if(u >= texture->width || u < GLFix(0))
                    u = 0;
                if(v >= texture->height || v < GLFix(0))
                    v = 0;

                pixel(x, y, z, texture->bitmap[u.to_int() + v.to_int()*texture->width]);
#else
                pixel(x, y, z, c);
#endif
            }
        }

        xstart += dx_far;
        zstart += dz_far;
        xend += dx_upper;
        zend += dz_upper;

#ifdef TEXTURE_SUPPORT
        ustart += du_far;
        vstart += dv_far;
        uend += du_upper;
        vend += dv_upper;
#endif
    }
}

void set_color(COLOR c)
{
	color = c;
}

static unsigned int nvertecies = 0;
static VERTEX vertecies[4];
static GLDrawMode draw_mode = GL_TRIANGLES;

void add_vertex(VERTEX* vertex)
{
    mult_mat_vect_res(transformation, vertex, &vertecies[nvertecies]);

    nvertecies++;

    GLFix x1, y1, x2, y2;

    switch(draw_mode)
    {
    case GL_TRIANGLES:
        if(nvertecies != 3)
            break;

        nvertecies = 0;

        //Backface culling
        x1 = vertecies[1].x - vertecies[0].x, x2 = vertecies[2].x - vertecies[0].x;
        y1 = vertecies[1].y - vertecies[0].y, y2 = vertecies[2].y - vertecies[0].y;

        if(y1 * x2 < y2 * x1)
        {
            if(true)
                draw_triangle(&vertecies[0], &vertecies[1], &vertecies[2], color);

            else
            {
                draw_line_3d(&vertecies[0], &vertecies[1], color);
                draw_line_3d(&vertecies[0], &vertecies[2], color);
                draw_line_3d(&vertecies[2], &vertecies[1], color);
            }
        }
        break;

    case GL_QUADS:
        if(nvertecies != 4)
            break;

        nvertecies = 0;

        //Backface culling: The quad must be on one plane!
        x1 = vertecies[1].x - vertecies[0].x, x2 = vertecies[2].x - vertecies[0].x;
        y1 = vertecies[1].y - vertecies[0].y, y2 = vertecies[2].y - vertecies[0].y;

        if(y1 * x2 < y2 * x1)
        {
            if(true)
            {
                draw_triangle(&vertecies[0], &vertecies[1], &vertecies[2], color);
                draw_triangle(&vertecies[2], &vertecies[3], &vertecies[0], color);
            }
            else
            {
                draw_line_3d(&vertecies[0], &vertecies[1], color);
                draw_line_3d(&vertecies[1], &vertecies[2], color);
                draw_line_3d(&vertecies[2], &vertecies[3], color);
                draw_line_3d(&vertecies[3], &vertecies[0], color);
            }
        }
        break;
    case GL_QUAD_STRIP:
        if(nvertecies != 4)
            break;

        nvertecies = 2;

        //Backface culling: The quad must be on one plane!
        x1 = vertecies[1].x - vertecies[0].x, x2 = vertecies[2].x - vertecies[0].x;
        y1 = vertecies[1].y - vertecies[0].y, y2 = vertecies[2].y - vertecies[0].y;

        if(y1 * x2 < y2 * x1)
        {
            if(true)
            {
                draw_triangle(&vertecies[0], &vertecies[1], &vertecies[2], color);
                draw_triangle(&vertecies[2], &vertecies[3], &vertecies[0], color);
            }
            else
            {
                draw_line_3d(&vertecies[0], &vertecies[1], color);
                draw_line_3d(&vertecies[1], &vertecies[2], color);
                draw_line_3d(&vertecies[2], &vertecies[3], color);
                draw_line_3d(&vertecies[3], &vertecies[0], color);
            }
        }

        vertecies[0] = vertecies[2];
        vertecies[1] = vertecies[3];
        break;
    }
}

static GLFix u, v;

void glBindTexture(TEXTURE *tex)
{
    texture = tex;
}

void glTexCoord2f(GLFix nu, GLFix nv)
{
    u = nu;
    v = nv;
}

void glVertex3f(GLFix x, GLFix y, GLFix z)
{
    VERTEX ver = {
			.x = x,
			.y = y,
            .z = z,
            .u = u,
            .v = v
	};

    add_vertex(&ver);
}

void glBegin(GLDrawMode mode)
{
    nvertecies = 0;
    draw_mode = mode;
}

//Could be faster than memset, but allows us to specify a 32-bit value instead of 8-bit
inline void memclear(uint32_t value, uint32_t *buffer, uint32_t size)
{
    while(size--)
        *buffer++ = value;
}

void glClear(int buffers)
{
    if(buffers & GL_COLOR_BUFFER_BIT)
        memclear(color | (color<<16), reinterpret_cast<uint32_t*>(screen), WIDTH*HEIGHT/2); //16-bit, so /2

    //The depth buffer uses unsigned 32-bit ints internally
    if(buffers & GL_DEPTH_BUFFER_BIT)
        memclear(UINT32_MAX, reinterpret_cast<uint32_t*>(z_buffer), WIDTH*HEIGHT);
}

void glColor3f(GLFix r, GLFix g, GLFix b)
{
    int r1 = (r * GLFix(0b11111)).round();
    int g1 = (g * GLFix(0b111111)).round();
    int b1 = (b * GLFix(0b11111)).round();

    COLOR c = ((r1 & 0b11111) << 11) | ((g1 & 0b111111) << 5) | (b1 & 0b11111);
	set_color(c);
}

void glPopMatrix()
{
    transformation--;
}

void glPushMatrix()
{
    MATRIX* n = transformation + 1;
    *n = *transformation;
    transformation = n;
}
